using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ConvexHulls
{
    class BruteForceHull : IConvexHull
    {
        /// <summary>
        /// Calculates the convex hull.
        /// </summary>
        public Point[] CalculateHull(Point[] points)
        {
            if (points.Length <= 2)
                return points;

            List<Point> hull = new List<Point>();
            // for every point look at every other point
            for (int i = 0; i < points.Length; i++)
            {
                Point p1 = points[i];
                for (int j = 0; j < points.Length; j++)
                {
                    Point p2 = points[j];
                    if (p2 != p1)
                    {
                        // check if line p1->p2 is part of the convex hull
                        bool isHullPart = true;
                        for (int k = 0; k < points.Length && isHullPart; k++)
                        {
                            Point p3 = points[k];
                            if (p1 != p3 && p2 != p3 && isPLeft(p3, p2, p1))
                                isHullPart = false;
                        }
                        if (isHullPart && !hull.Contains(p2))
                            hull.Add(p2);
                    }
                }
            }

            return hull.ToArray();
        }

        /// <summary>
        /// Returns the distance of a point p to a vector (from,to).
        /// </summary>
        private int getDist(Point p, Point vfrom, Point vto)
        {
            return ((vto.X - vfrom.X) * (p.Y - vfrom.Y) - (p.X - vfrom.X) * (vto.Y - vfrom.Y));
        }

        /// <summary>
        /// Returns true, if the point p is on the left hand side of the vector vfrom->vto.
        /// </summary>
        private bool isPLeft(Point p, Point vfrom, Point vto)
        {
            return (getDist(p, vfrom, vto) < 0);
        }
    }
}
