using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.IO;

namespace ConvexHulls
{
    class StatisticExperiments
    {
        int maxn = 1000000;
        int diffn = 50000;
        int runs = 50;
        int maxval = (int)Math.Sqrt(Int32.MaxValue);   // maxval to avoid integer overflows


        // timer for time measure
        HiResTimer timer = new HiResTimer();

        /// <summary>
        /// Runs the defined experiments.
        /// </summary>
        public void Run()
        {
            System.Console.WriteLine("Starte Auswertung...");

            long stime = timer.Value;
            expMain();
            expBruteForce();

            long etime = timer.Value;
            System.Console.WriteLine("Gesamtzeit: " + ((etime - stime) / (double)timer.Frequency) / 3600 + "h");

            System.Console.WriteLine("Auswertung beendet...");
        }

        private void expMain()
        {
            System.Console.WriteLine("\n ===>  Starte Auswertung expMain()  <===");
            Random rnd = new Random();
            int steps = maxn / diffn;

            int algs = 4;
            IConvexHull[] hullAlgs = new IConvexHull[algs];
            hullAlgs[0] = new JarvisMarch();
            hullAlgs[1] = new GrahamScan();
            hullAlgs[2] = new QuickHull();
            hullAlgs[3] = new ChanHull();
            string[] algNames = new string[algs];
            algNames[0] = "JarvisMarch";
            algNames[1] = "GrahamScan";
            algNames[2] = "QuickHull";
            algNames[3] = "ChanHull";

            int exps = 3;
            string[] expNames = {"Random", "Circle", "Square"};

            // without and with Akl Toussaint Heuristics
            for (int a = 0; a < 2; a++)
            {
                System.Console.WriteLine("\n\n\n ====> " + (a == 0 ? "OHNE" : "MIT") + " Akl Toussaint Heuristik <====\n\n");
                // run the different experiments
                for (int b = 0; b < exps; b++)
                {
                    System.Console.WriteLine("\n----> Starte Experiment: " + expNames[b] + " <----");
                    // time measurement
                    double[,] itimes = new double[algs, steps];
                    for (int i = 0, n = diffn; i < steps; i++, n += diffn)
                    {
                        System.Console.WriteLine("=> " + n + " Elemente");
                        System.Console.Write("   Durchlauf: ");
                        for (int k = 0; k < runs; k++)
                        {
                            System.Console.Write((k + 1) + ", ");
                            // create point set
                            Point[] orig = null;
                            if (b == 0)
                                orig = createRandomly(n);
                            else if (b == 1)
                                orig = createCircle(n);
                            else
                                orig = createSquare(n);
                            // run experiments for all algs
                            for (int j = 0; j < algs; j++)
                            {
                                if (!(b == 1 && algNames[j] == "JarvisMarch"))
                                {
                                    Point[] points = (Point[])orig.Clone();
                                    long stime = timer.Value;
                                    if (a == 1)
                                        points = AklToussaintHeuristic.PrunePoints(points);
                                    Point[] hull = hullAlgs[j].CalculateHull(points);
                                    long etime = timer.Value;
                                    itimes[j, i] += (etime - stime) / (double)timer.Frequency;
                                }
                            }
                        }
                        System.Console.WriteLine();
                        for (int j = 0; j < algs; j++)
                        {
                            itimes[j, i] /= runs;
                            System.Console.WriteLine("- " + algNames[j] + ": " + itimes[j, i] + "s");
                        }
                    }

                    // write in gnuplot file
                    string fname = "Hulls_Time_" + expNames[b] + (a == 0 ? "" : "_AklHeur") + ".txt";
                    StreamWriter otimes = new StreamWriter(fname);
                    for (int i = 0; i < algs; i++)
                    {
                        otimes.WriteLine("# " + algNames[i]);

                        for (int j = 0, n = diffn; j < steps; j++, n += diffn)
                        {
                            otimes.WriteLine(n + " " + itimes[i, j].ToString().Replace(',', '.'));
                        }

                        otimes.WriteLine();
                        otimes.WriteLine();
                    }
                    otimes.Close();
                }
            }


            // repeat Jarvis March for circle
            maxn = 10000;
            diffn = 500;
            steps = maxn / diffn;
            for (int a = 0; a < 2; a++)
            {
                System.Console.WriteLine("\n----> Starte Experiment: Circle JV <----");
                // time measurement
                double[,] itimes = new double[algs, steps];
                for (int i = 0, n = diffn; i < steps; i++, n += diffn)
                {
                    System.Console.WriteLine("=> " + n + " Elemente");
                    System.Console.Write("   Durchlauf: ");
                    for (int k = 0; k < runs; k++)
                    {
                        System.Console.Write((k + 1) + ", ");
                        // create point set
                        Point[] orig = createCircle(n);
                        // run experiments for all algs
                        for (int j = 0; j < algs; j++)
                        {
                            Point[] points = (Point[])orig.Clone();
                            long stime = timer.Value;
                            if (a == 1)
                                points = AklToussaintHeuristic.PrunePoints(points);
                            Point[] hull = hullAlgs[j].CalculateHull(points);
                            long etime = timer.Value;
                            itimes[j, i] += (etime - stime) / (double)timer.Frequency;
                        }
                    }
                    System.Console.WriteLine();
                    for (int j = 0; j < algs; j++)
                    {
                        itimes[j, i] /= runs;
                        System.Console.WriteLine("- " + algNames[j] + ": " + itimes[j, i] + "s");
                    }
                }

                // write in gnuplot file
                string fname = "Hulls_Time_JV_Circle" + (a == 0 ? "" : "_AklHeur") + ".txt";
                StreamWriter otimes = new StreamWriter(fname);
                for (int i = 0; i < algs; i++)
                {
                    otimes.WriteLine("# " + algNames[i]);

                    for (int j = 0, n = diffn; j < steps; j++, n += diffn)
                    {
                        otimes.WriteLine(n + " " + itimes[i, j].ToString().Replace(',', '.'));
                    }

                    otimes.WriteLine();
                    otimes.WriteLine();
                }
                otimes.Close();
            }
        }

        private void expBruteForce()
        {
            System.Console.WriteLine("\n ===>  Starte Auswertung expBruteForce()  <===");
            Random rnd = new Random();
            
            int algs = 3;
            IConvexHull[] hullAlgs = new IConvexHull[algs];
            hullAlgs[0] = new BruteForceHull();
            hullAlgs[1] = new JarvisMarch();
            hullAlgs[2] = new QuickHull();
            string[] algNames = new string[algs];
            algNames[0] = "BruteForceHull";
            algNames[1] = "JarvisMarch";
            algNames[2] = "QuickHull";

            int exps = 3;
            string[] expNames = { "Random", "Circle", "Square" };

            // without and with Akl Toussaint Heuristics
            for (int a = 0; a < 2; a++)
            {
                System.Console.WriteLine("\n\n\n ====> " + (a == 0 ? "OHNE" : "MIT") + " Akl Toussaint Heuristik <====\n\n");
                // run the different experiments
                for (int b = 0; b < exps; b++)
                {
                    int maxn = 2500;
                    int diffn = 125;
                    if (b == 1)
                    {
                        maxn = 500;
                        diffn = 25;
                    }
                    int steps = maxn / diffn;

                    System.Console.WriteLine("\n----> Starte Experiment: " + expNames[b] + " <----");
                    // time measurement
                    double[,] itimes = new double[algs, steps];
                    for (int i = 0, n = diffn; i < steps; i++, n += diffn)
                    {
                        System.Console.WriteLine("=> " + n + " Elemente");
                        System.Console.Write("   Durchlauf: ");
                        for (int k = 0; k < runs; k++)
                        {
                            System.Console.Write((k + 1) + ", ");
                            // create point set
                            Point[] orig = null;
                            if (b == 0)
                                orig = createRandomly(n);
                            else if (b == 1)
                                orig = createCircle(n);
                            else
                                orig = createSquare(n);
                            // run experiments for all algs
                            for (int j = 0; j < algs; j++)
                            {
                                Point[] points = (Point[])orig.Clone();
                                long stime = timer.Value;
                                if (a == 1)
                                    points = AklToussaintHeuristic.PrunePoints(points);
                                Point[] hull = hullAlgs[j].CalculateHull(points);
                                long etime = timer.Value;
                                itimes[j, i] += (etime - stime) / (double)timer.Frequency;
                            }
                        }
                        System.Console.WriteLine();
                        for (int j = 0; j < algs; j++)
                        {
                            itimes[j, i] /= runs;
                            System.Console.WriteLine("- " + algNames[j] + ": " + itimes[j, i] + "s");
                        }
                    }

                    // write in gnuplot file
                    string fname = "Hulls_Time_BF_" + expNames[b] + (a == 0 ? "" : "_AklHeur") + ".txt";
                    StreamWriter otimes = new StreamWriter(fname);
                    for (int i = 0; i < algs; i++)
                    {
                        otimes.WriteLine("# " + algNames[i]);

                        for (int j = 0, n = diffn; j < steps; j++, n += diffn)
                        {
                            otimes.WriteLine(n + " " + itimes[i, j].ToString().Replace(',', '.'));
                        }

                        otimes.WriteLine();
                        otimes.WriteLine();
                    }
                    otimes.Close();
                }
            }
        }

        private void expCircle()
        {

        }

        private void expSquare()
        {

        }

        private Point[] createCircle(int n)
        {
            int r = maxval / 2 - 1;
            Point[] points = new Point[n];
            for (int i = 0; i < n; i++)
            {
                double p = 2 * i * Math.PI / n;
                points[i] = new Point((int)(Math.Cos(p) * r + r), (int)(Math.Sin(p) * r + r));
            }
            return points;
        }

        private Point[] createSquare(int n)
        {
            Point[] points = new Point[n];
            points[0] = new Point(0, 0);
            points[1] = new Point(0, maxval);
            points[2] = new Point(maxval, maxval);
            points[3] = new Point(maxval, 0);
            Random rnd = new Random();
            for (int i = 4; i < n; i++)
            {
                points[i] = new Point(rnd.Next(maxval), rnd.Next(maxval));
            }
            return points;
        }

        private Point[] createRandomly(int n)
        {
            Point[] points = new Point[n];
            Random rnd = new Random();
            for (int i = 0; i < n; i++)
            {
                points[i] = new Point(rnd.Next(maxval), rnd.Next(maxval));
            }
            return points;
        }
    }
}
