using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Labyrinth
{
    class Room
    {
        /// <summary>
        /// True for a wall in this direction, false if there is no wall
        /// </summary>
        public bool North, East, South, West;

        private byte mark = 0;   // mark for visited rooms

        public byte Mark
        {
            get { return mark; }
        }

        public Room() { }

        public Room(bool north, bool east, bool south, bool west)
        {
            this.North = north;
            this.South = south;
            this.East = east;
            this.West = west;
        }

        public override string ToString()
        {
            string s = "Room [ ";
            if (North) s += "North ";
            if (East) s += "East ";
            if (South) s += "South ";
            if (West) s += "West ";
            s += "]";
            return s;
        }

        /// <summary>
        /// Set room to not visited.
        /// </summary>
        public void SetUnvisited()
        {
            mark = 0;
        }

        /// <summary>
        /// Set room to visited.
        /// </summary>
        public void SetVisited()
        {
            mark = 1;
        }

        /// <summary>
        /// Set room to visited, but not element of the current path through the labyrinth.
        /// </summary>
        public void SetVisitedNoPath()
        {
            mark = 2;
        }

        public bool IsUnvisited()
        {
            return (mark == 0);
        }

        public bool IsVisited()
        {
            return (mark == 1);
        }

        public bool IsVisitedNoPath()
        {
            return (mark == 2);
        }

        /// <summary>
        /// Returns a brush depending on the value of the mark.
        /// </summary>
        /// <returns></returns>
        public Brush GetBrush()
        {
            switch (mark)
            {
                case 1:   // room visited and part of current path
                    return Brushes.Red;
                case 2:   // room visited, but not part of path
                    return Brushes.Olive;
            }

            // room not visited
            return Brushes.White;
        }
    }
}
