package algorithms.priorityqueues;

import java.util.Stack;


public class BinomialHeapNode<K,V> implements PriorityQueueNode<K,V> {

	int priority = 0;
	int degree = 0;
	V value = null;
	K key = null;
	
	BinomialHeapNode<K,V> parent = null;
	BinomialHeapNode<K,V> sibling = null;
	BinomialHeapNode<K,V> firstChild = null;
	
	public BinomialHeapNode(V val, K key, int priority){
		this.value = val;
		this.key = key;
		this.priority = priority;
	}
	
	public BinomialHeapNode(V val, K key, int priority, int degree){
		this(val, key, priority);
		this.degree = degree;
	}
	
	public BinomialHeapNode(V val, K key, int priority, int degree, BinomialHeapNode<K,V> parent){
		this(val, key, priority, degree);
		this.parent = parent;
	}
	
	public BinomialHeapNode(V val, K key, int priority, int degree, BinomialHeapNode<K,V> parent, BinomialHeapNode<K,V> sibling){
		this(val, key, priority, degree, parent);
		this.sibling = sibling;
	}
	
	public int getPriority() {
		return priority;
	}

	public void setPriority(int val) {
		this.priority = val;
	}
	
	public K getKey(){
		return key;
	}
	
	public void setKey(K key){
		this.key = key;
	}
	
	public V getValue() {
		return value;
	}

	public void setValue(V val) {
		this.value = val;
	}

	public int getDegree() {
		return degree;
	}

	public void setDegree(int val) {
		this.degree = val;
	}

	public BinomialHeapNode<K,V> getParent() {
		return parent;
	}

	public void setParent(BinomialHeapNode<K,V> parent) {
		this.parent = parent;
	}

	public BinomialHeapNode<K,V> getSibling() {
		return sibling;
	}

	public void setSibling(BinomialHeapNode<K,V> sibling) {
		this.sibling = sibling;
	}

	public BinomialHeapNode<K,V> getFirstChild() {
		return firstChild;
	}

	public void setFirstChild(BinomialHeapNode<K,V> child) {
		this.firstChild = child;
	}
	
	
	public void linkWith(BinomialHeapNode<K,V> node){
		this.parent = node;
		this.sibling = node.firstChild;
		node.firstChild = this;
		node.degree++;
	}
	
	public void invertChildList(){
		BinomialHeapNode<K,V> node = firstChild;
		Stack<BinomialHeapNode<K,V>> children = new Stack<BinomialHeapNode<K,V>>();
		
		// get list of children in inverted order
		while(null != node){
			children.push(node);
			node = node.getSibling();
		}
		
		// set inverted list as child list
		if( !children.isEmpty() ){
			BinomialHeapNode<K,V> first = children.pop();
			node = first;
			while( !children.isEmpty() ){
				BinomialHeapNode<K,V> popped = children.pop();
				node.setSibling(popped);
				node = popped;
			}
			node.setSibling(null);
			firstChild = first;
		}
	}
	
	public void swapValues(BinomialHeapNode<K,V> n){
		// temp savings
		int prion = n.getPriority();
		K keyn = n.getKey();
		V valn = n.getValue();
		//int degn = n.getDegree();
		
		// set this values to n
		n.setPriority(this.priority);
		n.setKey(this.key);
		n.setValue(this.value);
		//n.setDegree(this.degree);
		
		// set temp values to this
		this.priority = prion;
		this.key = keyn;
		this.value = valn;
		//this.degree = degn;
	}

}
