package mains;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Random;

import algorithms.sorting.SortingAlgorithmsInt;

public class Excercise1 {
	private static int runs = 50;
	
	// measuring of all quicksort variants up to 4 mio. elements
	private static void quicksortTests() throws IOException{
		ThreadMXBean timer = ManagementFactory.getThreadMXBean();
		timer.setThreadCpuTimeEnabled(true);
		SortingAlgorithmsInt sorter = new SortingAlgorithmsInt();
		
		int n0 = 4000000;
		int div = 1000000;
		int mpoints = n0 / div;
		int m = 15;
		
		double[][] mtime = new double[mpoints][6];
		double[][] mops = new double[mpoints][6];
		double[][] mcompares = new double[mpoints][6];
		System.out.println("\n==> Starting quicksortTests()");
		// iteration over the element number, until maximum is reached
		for(int n=div, l=0; n<=n0; n+=div, l++){
			System.out.println("\n-> Elementzahl: " + n);
			// make several runs, calculate mean of those afterwards
			for(int i=1; i<=runs; i++){
				System.out.print(i+",");
				System.out.flush();
				int[] b = randomIntArray(n, n*10);
				
				// normal quicksort
				int[] a = b.clone();
				long ts = timer.getCurrentThreadCpuTime();
				sorter.qsort(a);
				long te = timer.getCurrentThreadCpuTime();
				mtime[l][0] += (te-ts)*Math.pow(10,-9);
				mops[l][0] += sorter.getOps();
				mcompares[l][0] += sorter.getCompares();
				sorter.resetCounters();
				
				// iterative qsort combined with insertion sort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortItISort(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][1] += (te-ts)*Math.pow(10,-9);
				mops[l][1] += sorter.getOps();
				mcompares[l][1] += sorter.getCompares();
				sorter.resetCounters();
				
				// iterative qsort only 
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortIt(a);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][2] += (te-ts)*Math.pow(10,-9);
				mops[l][2] += sorter.getOps();
				mcompares[l][2] += sorter.getCompares();
				sorter.resetCounters();
				
				// recursive qsort combined with insertion sort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortISort(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][3] += (te-ts)*Math.pow(10,-9);
				mops[l][3] += sorter.getOps();
				mcompares[l][3] += sorter.getCompares();
				sorter.resetCounters();
				
				// recursive qsort combined with median-of-three
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortMo3(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][4] += (te-ts)*Math.pow(10,-9);
				mops[l][4] += sorter.getOps();
				mcompares[l][4] += sorter.getCompares();
				sorter.resetCounters();
				
				// recursive qsort combined with median-of-three and insertion sort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortMo3ISort(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][5] += (te-ts)*Math.pow(10,-9);
				mops[l][5] += sorter.getOps();
				mcompares[l][5] += sorter.getCompares();
				sorter.resetCounters();
			}
			System.out.println();
			for(int i=0; i<6; i++){
				mtime[l][i] /= runs;
				mops[l][i] /= runs;
				mcompares[l][i] /= runs;
				System.out.println("i="+i+": time="+mtime[l][i]+"s, ops="+mops[l][i]+", compares="+mcompares[l][i]);
			}
		}
		
		// write the results into a file
		Writer ostime = new FileWriter("QuickSortsCmpTime.txt");
		Writer osops = new FileWriter("QuickSortsCmpOps.txt");
		Writer oscompares = new FileWriter("QuickSortsCmpCompares.txt");
		for(int i=0; i<6; i++){
			switch(i){
			case 0:
				ostime.write("# Quicksort\n");
				osops.write("# Quicksort\n");
				oscompares.write("# Quicksort\n");
				break;
			case 1:
				ostime.write("# Tuned Quicksort (iterative qsort + insertion sort)\n");
				osops.write("# Tuned Quicksort (iterative qsort + insertion sort)\n");
				oscompares.write("# Tuned Quicksort (iterative qsort + insertion sort)\n");
				break;
			case 2:
				ostime.write("# Tuned Quicksort (iterative qsort)\n");
				osops.write("# Tuned Quicksort (iterative qsort)\n");
				oscompares.write("# Tuned Quicksort (iterative qsort)\n");
				break;
			case 3:
				ostime.write("# Tuned Quicksort (recursive qsort + insertion sort)\n");
				osops.write("# Tuned Quicksort (recursive qsort + insertion sort)\n");
				oscompares.write("# Tuned Quicksort (recursive qsort + insertion sort)\n");
				break;
			case 4:
				ostime.write("# Tuned Quicksort (recursive qsort + median-of-three)\n");
				osops.write("# Tuned Quicksort (recursive qsort + median-of-three)\n");
				oscompares.write("# Tuned Quicksort (recursive qsort + median-of-three)\n");
				break;
			case 5:
				ostime.write("# Tuned Quicksort (recursive qsort + median-of-three + insertion sort)\n");
				osops.write("# Tuned Quicksort (recursive qsort + median-of-three + insertion sort)\n");
				oscompares.write("# Tuned Quicksort (recursive qsort + median-of-three + insertion sort)\n");
				break;
			}
			ostime.write("0 0\n");
			osops.write("0 0\n");
			oscompares.write("0 0\n");
			for(int j=0; j<mtime.length; j++){
				ostime.write(div+j*div + " " + mtime[j][i] + "\n");
				osops.write(div+j*div + " " + mops[j][i] + "\n");
				oscompares.write(div+j*div + " " + mcompares[j][i] + "\n");
			}
			ostime.write("\n\n");
			osops.write("\n\n");
			oscompares.write("\n\n");
		}
		ostime.close();
		osops.close();
		oscompares.close();
	}
	
	// qsort, tuned qsort and heapsort in comparison up to 4 mio. elements
	public static void quickHeapSortTests() throws IOException{
		ThreadMXBean timer = ManagementFactory.getThreadMXBean();
		timer.setThreadCpuTimeEnabled(true);
		SortingAlgorithmsInt sorter = new SortingAlgorithmsInt();
		
		int n0 = 4000000;
		int div = 50000;
		int mpoints = n0 / div;
		int m = 23;
		
		double[][] mtime = new double[mpoints][4];
		double[][] mops = new double[mpoints][4];
		double[][] mcompares = new double[mpoints][4];
		System.out.println("\n==> Starting quickHeapSortTests()");
		// iteration over the element number in the arrays to sort
		for(int n=div, l=0; n<=n0; n+=div, l++){
			System.out.println("\n-> Elementzahl: " + n);
			// make several runs, calculate mean of those afterwards
			for(int i=1; i<=runs; i++){
				System.out.print(i+",");
				System.out.flush();
				int[] b = randomIntArray(n, n*10);

				// normal quicksort
				int[] a = b.clone();
				long ts = timer.getCurrentThreadCpuTime();
				sorter.qsort(a);
				long te = timer.getCurrentThreadCpuTime();
				mtime[l][0] += (te-ts)*Math.pow(10,-9);
				mops[l][0] += sorter.getOps();
				mcompares[l][0] += sorter.getCompares();
				sorter.resetCounters();
				
				// tuned qsort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortMo3ISort(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][1] += (te-ts)*Math.pow(10,-9);
				mops[l][1] += sorter.getOps();
				mcompares[l][1] += sorter.getCompares();
				sorter.resetCounters();
				
				// normal heapsort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.heapsort(a);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][2] += (te-ts)*Math.pow(10,-9);
				mops[l][2] += sorter.getOps();
				mcompares[l][2] += sorter.getCompares();
				sorter.resetCounters();
				
				// Bottom-Up-Heapsort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.bottomUpHeapsort(a);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][3] += (te-ts)*Math.pow(10,-9);
				mops[l][3] += sorter.getOps();
				mcompares[l][3] += sorter.getCompares();
				sorter.resetCounters();
			}
			System.out.println();
			for(int i=0; i<4; i++){
				mtime[l][i] /= runs;
				mops[l][i] /= runs;
				mcompares[l][i] /= runs;
				System.out.println("i="+i+": time="+mtime[l][i]+"s, ops="+mops[l][i]+", compares="+mcompares[l][i]);
			}
		}
		
		// write the results into a file
		Writer ostime = new FileWriter("QuickHeapCmpTime.txt");
		Writer osops = new FileWriter("QuickHeapCmpOps.txt");
		Writer oscompares = new FileWriter("QuickHeapCmpCompares.txt");
		for(int i=0; i<4; i++){
			switch(i){
			case 0:
				ostime.write("# Quicksort\n");
				osops.write("# Quicksort\n");
				oscompares.write("# Quicksort\n");
				break;
			case 1:
				ostime.write("# Tuned Quicksort\n");
				osops.write("# Tuned Quicksort\n");
				oscompares.write("# Tuned Quicksort\n");
				break;
			case 2:
				ostime.write("# Heapsort\n");
				osops.write("# Heapsort\n");
				oscompares.write("# Heapsort\n");
				break;
			case 3:
				ostime.write("# Bottom-Up Heapsort\n");
				osops.write("# Bottom-Up Heapsort\n");
				oscompares.write("# Bottom-Up Heapsort\n");
				break;
			}
			ostime.write("0 0\n");
			osops.write("0 0\n");
			oscompares.write("0 0\n");
			for(int j=0; j<mtime.length; j++){
				ostime.write(div+j*div + " " + mtime[j][i] + "\n");
				osops.write(div+j*div + " " + mops[j][i] + "\n");
				oscompares.write(div+j*div + " " + mcompares[j][i] + "\n");
			}
			ostime.write("\n\n");
			osops.write("\n\n");
			oscompares.write("\n\n");
		}
		ostime.close();
		osops.close();
		oscompares.close();
	}
	
	// find with tuned quicksort the optimal switch size to insertion sort
	public static void findOptimalISortBreakPoint() throws IOException{
		ThreadMXBean timer = ManagementFactory.getThreadMXBean();
		timer.setThreadCpuTimeEnabled(true);
		SortingAlgorithmsInt sorter = new SortingAlgorithmsInt();
		
		int[] n0 = new int[]{1000000, 3000000, 5000000, 7000000};
		int mmax = 50;
		int div = 1;
		int mpoints = mmax / div;
		
		double[][] mtime = new double[n0.length][mpoints];
		double[][] mops = new double[n0.length][mpoints];
		double[][] mcompares = new double[n0.length][mpoints];
		System.out.println("\n==> Starting findOptimalISortBreakPoint()");
		// do experiments for all sizes of the arrays to sort
		for(int k=0; k<n0.length; k++){
			int n = n0[k];
			System.out.println("\n-> Elementzahl: " + n);
			// iterate over the switch size m up to mmax
			for(int m=div, l=0; m<=mmax; m+=div, l++){
				System.out.println("-  m = " + m);
				// make several runs, calculate mean of those afterwards
				for(int i=1; i<=runs; i++){
					System.out.print(i+",");
					System.out.flush();
					int[] b = randomIntArray(n, n*10);
					// sort with tuned quicksort 
					int[] a = b.clone();
					long ts = timer.getCurrentThreadCpuTime();
					sorter.qsortMo3ISort(a, m);
					long te = timer.getCurrentThreadCpuTime();
					mtime[k][l] += (te-ts)*Math.pow(10,-9);
					mops[k][l] += sorter.getOps();
					mcompares[k][l] += sorter.getCompares();
					sorter.resetCounters();
				}
				System.out.println();
				mtime[k][l] /= runs;
				mops[k][l] /= runs;
				mcompares[k][l] /= runs;
				System.out.println("time="+mtime[k][l]+"s");
			}
		}
		
		// write the results into a file
		Writer ostime = new FileWriter("QuickISortCmpTime.txt");
		Writer osops = new FileWriter("QuickISortCmpOps.txt");
		Writer oscompares = new FileWriter("QuickISortCmpCompares.txt");
		for(int i=0; i<n0.length; i++){
			ostime.write("# Tuned Quicksort, "+n0[i]+" elements\n");
			for(int j=0; j<mtime[0].length; j++){
				ostime.write(div+j*div + " " + mtime[i][j] + "\n");
				osops.write(div+j*div + " " + mops[i][j] + "\n");
				oscompares.write(div+j*div + " " + mcompares[i][j] + "\n");
			}
			ostime.write("\n\n");
			osops.write("\n\n");
			oscompares.write("\n\n");
		}
		ostime.close();
		osops.close();
		oscompares.close();
	}
	
	// comparison of quicksort and heapsort for 1 mio. elements with artificial stretching
	// of the comparison times for two elements
	public static void findQuickHeapCrossTimes() throws IOException{
		ThreadMXBean timer = ManagementFactory.getThreadMXBean();
		timer.setThreadCpuTimeEnabled(true);
		SortingAlgorithmsInt sorter = new SortingAlgorithmsInt();
		
		int n0 = 1000000;
		int maxIts = 1000;
		int div = 50;
		int mpoints = maxIts / div;
		int m = 15;
		
		double[][] mtime = new double[mpoints+1][3];
		double[][] mops = new double[mpoints+1][3];
		double[][] mcompares = new double[mpoints+1][3];
		System.out.println("\n==> Starting findQuickHeapCrossTimes()");
		// iterate over the number of iterations for artificial stretching of the comparison time
		for(int its=0, l=0; its<=maxIts; its+=div, l++){
			sorter.setCmpDelayIts(its);
			System.out.println("\n-> Delay iterations: " + its);
			// make several runs, calculate mean of those afterwards
			for(int i=1; i<=runs; i++){
				System.out.print(i+",");
				System.out.flush();
				int[] b = randomIntArray(n0, n0*10);
				
				// normal quicksort
				int[] a = b.clone();
				long ts = timer.getCurrentThreadCpuTime();
				sorter.qsort(a);
				long te = timer.getCurrentThreadCpuTime();
				mtime[l][0] += (te-ts)*Math.pow(10,-9);
				mops[l][0] += sorter.getOps();
				mcompares[l][0] += sorter.getCompares();
				sorter.resetCounters();
				
				// tuned quicksort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.qsortMo3ISort(a, m);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][1] += (te-ts)*Math.pow(10,-9);
				mops[l][0] += sorter.getOps();
				mcompares[l][0] += sorter.getCompares();
				sorter.resetCounters();
				
				// Bottom-Up-Heapsort
				a = b.clone();
				ts = timer.getCurrentThreadCpuTime();
				sorter.bottomUpHeapsort(a);
				te = timer.getCurrentThreadCpuTime();
				mtime[l][2] += (te-ts)*Math.pow(10,-9);
				mops[l][0] += sorter.getOps();
				mcompares[l][0] += sorter.getCompares();
				sorter.resetCounters();
			}
			sorter.resetCmpDelayIts();
			System.out.println();
			for(int i=0; i<3; i++){
				mtime[l][i] /= runs;
				mops[l][i] /= runs;
				mcompares[l][i] /= runs;
				System.out.println("i="+i+": time="+mtime[l][i]+"s");
			}
		}
		
		// write the results into a file
		Writer ostime = new FileWriter("QuickHeapDelayCmpTime.txt");
		Writer osops = new FileWriter("QuickHeapDelayCmpOps.txt");
		Writer oscompares = new FileWriter("QuickHeapDelayCmpCompares.txt");
		for(int i=0; i<3; i++){
			switch(i){
			case 0:
				ostime.write("# Quicksort\n");
				osops.write("# Quicksort\n");
				oscompares.write("# Quicksort\n");
				break;
			case 1:
				ostime.write("# Tuned Quicksort\n");
				osops.write("# Tuned Quicksort\n");
				oscompares.write("# Tuned Quicksort\n");
				break;
			case 2:
				ostime.write("# Bottom-Up Heapsort\n");
				osops.write("# Bottom-Up Heapsort\n");
				oscompares.write("# Bottom-Up Heapsort\n");
				break;
			}
			for(int j=0; j<mtime.length; j++){
				ostime.write(j*div + " " + mtime[j][i] + "\n");
				osops.write(j*div + " " + mops[j][i] + "\n");
				oscompares.write(j*div + " " + mcompares[j][i] + "\n");
			}
			ostime.write("\n\n");
			osops.write("\n\n");
			oscompares.write("\n\n");
		}
		ostime.close();
		osops.close();
		oscompares.close();
	}
	
	// creates a randomly generated int-array of size n with elements in interval [0..max]
	private static int[] randomIntArray(int n, int max){
		Random rnd = new Random();
		int[] a = new int[n];
		for(int i=0; i<a.length; i++)
			a[i] = rnd.nextInt(max+1);
		return a;
	}
	
	public static void main(String[] args) {
		// run all tests
		try {
			quicksortTests();
			quickHeapSortTests();
			findOptimalISortBreakPoint();
			findQuickHeapCrossTimes();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
